package be.iminds.ilabt.jfed.experimenter_gui.dialogs;

import be.iminds.ilabt.jfed.experimenter_gui.config.JFedGuiConfig;
import be.iminds.ilabt.jfed.experimenter_gui.events.OpenESpecEvent;
import be.iminds.ilabt.jfed.experimenter_gui.ext_requests.JFedEnvironment;
import be.iminds.ilabt.jfed.experimenter_gui.util.BrowserUtil;
import be.iminds.ilabt.jfed.git.GitFetcherCommandBuilder;
import be.iminds.ilabt.jfed.highlevel.model.SfaModel;
import be.iminds.ilabt.jfed.highlevel.tasks.HighLevelTaskFactory;
import be.iminds.ilabt.jfed.lowlevel.user.GeniUserProvider;
import be.iminds.ilabt.jfed.gui_preferences.GuiPreferenceKey;
import be.iminds.ilabt.jfed.gui_preferences.JFedGuiPreferences;
import javafx.beans.binding.Bindings;
import javafx.beans.binding.BooleanBinding;
import javafx.beans.binding.ObjectBinding;
import javafx.event.ActionEvent;
import javafx.event.EventHandler;
import javafx.fxml.FXML;
import javafx.scene.control.*;
import javafx.scene.layout.StackPane;
import javafx.stage.DirectoryChooser;
import javafx.stage.FileChooser;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.inject.Inject;
import java.io.File;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.Optional;

/**
 * User: twalcari
 * Date: 2/24/14
 * Time: 4:03 PM
 */
public class OpenESpecController extends StackPane {
    private static final Logger LOG = LoggerFactory.getLogger(OpenESpecController.class);

    @Nonnull
    private final HighLevelTaskFactory hltf;
    @Nonnull
    private final GeniUserProvider geniUserProvider;
    @Nonnull
    private final SfaModel sfaModel;
    @Nonnull
    private final JFedEnvironment environment;
    @Nonnull
    private final JFedGuiConfig config;
    @Nonnull
    private final JFedGuiPreferences preferences;
    @Nonnull
    private final JFedGuiConfig guiConfig;

    private EventHandler<ActionEvent> actionEventHandler;

    private ObjectBinding<OpenESpecEvent.ESpecSource> especSource;
    private BooleanBinding valid;


    @Nonnull
    private final BrowserUtil browserUtil;

    @FXML
    protected ToggleGroup sourceToggleGroup;

    @FXML
    protected RadioButton archiveSourceButton;
    @FXML
    protected RadioButton dirSourceButton;
    @FXML
    protected RadioButton urlSourceButton;
    @FXML
    protected RadioButton gitSourceButton;

    @FXML
    protected TextField openArchiveTextField;
    @FXML
    protected TextField openDirTextField;
    @FXML
    protected TextField openURLTextField;
    @FXML
    protected TextField openGitRepoTextField;
    @FXML
    protected TextField openGitDirTextField;
    @FXML
    protected TextField openGitBranchTextField;


    @FXML
    protected CheckBox useGitDirCheckBox;
    @FXML
    protected CheckBox useGitBranchCheckBox;


    @FXML
    private Button especDocumentationButton;

    @Inject
    OpenESpecController(@Nonnull HighLevelTaskFactory hltf,
                        @Nonnull GeniUserProvider geniUserProvider,
                        @Nonnull SfaModel sfaModel,
                        @Nonnull JFedEnvironment environment,
                        @Nonnull JFedGuiConfig config,
                        @Nonnull JFedGuiPreferences preferences,
                        @Nonnull JFedGuiConfig guiConfig,
                        @Nonnull BrowserUtil browserUtil) {
        this.hltf = hltf;
        this.geniUserProvider = geniUserProvider;
        this.sfaModel = sfaModel;
        this.environment = environment;
        this.config = config;
        this.preferences = preferences;
        this.guiConfig = guiConfig;
        this.browserUtil = browserUtil;
    }


    @FXML
    public void initialize() {
        EventHandler<ActionEvent> eventPasser =
                event -> {
                    if (actionEventHandler != null) actionEventHandler.handle(event);
                };
        openArchiveTextField.setOnAction(eventPasser);
        openDirTextField.setOnAction(eventPasser);
        openURLTextField.setOnAction(eventPasser);
        openGitRepoTextField.setOnAction(eventPasser);

        String initialEspec = preferences.getString(GuiPreferenceKey.PREF_LAST_OPENED_ESPEC);
        if (initialEspec == null)
            initialEspec = guiConfig.getStringFlag("defaultESpecUrl");
        if (initialEspec != null && !initialEspec.trim().isEmpty()) {
            if (initialEspec.startsWith("git ") || initialEspec.startsWith("github ")) {
                gitSourceButton.setSelected(true);
                Optional<GitFetcherCommandBuilder> help = GitFetcherCommandBuilder.fromString(initialEspec);
                if (help.isPresent()) {
                    openGitRepoTextField.setText(help.get().getGitUrl());

                    if (help.get().getBranch() != null && !help.get().getBranch().trim().isEmpty()) {
                        useGitBranchCheckBox.setSelected(true);
                        openGitBranchTextField.setText(help.get().getBranch());
                    } else {
                        useGitBranchCheckBox.setSelected(false);
                    }

                    if (help.get().getRepoSubDir() != null && !help.get().getRepoSubDir().trim().isEmpty()) {
                        useGitDirCheckBox.setSelected(true);
                        openGitDirTextField.setText(help.get().getRepoSubDir());
                    } else {
                        useGitDirCheckBox.setSelected(false);
                    }
                }
            } else {
                if (initialEspec.startsWith("http")) {
                    urlSourceButton.setSelected(true);
                    openURLTextField.setText(initialEspec);
                } else {
                    if (initialEspec.endsWith("/") || new File(initialEspec).isDirectory()) {
                        dirSourceButton.setSelected(true);
                        openDirTextField.setText(initialEspec);
                    } else {
                        archiveSourceButton.setSelected(true);
                        openArchiveTextField.setText(initialEspec);
                    }
                }
            }
        }

        especDocumentationButton.setOnAction(event -> {
            //show in console, as fallback
            LOG.info("Clicked documentation button to ESpec documentation at https://jfed.ilabt.imec.be/espec/");

            //(try to) open in browser
            browserUtil.openUrlInBrowser("https://jfed.ilabt.imec.be/espec/");
        });

        especSource = Bindings.createObjectBinding(this::getESpecSource, sourceToggleGroup.selectedToggleProperty());

        BooleanBinding archiveBooleanBinding = Bindings.createBooleanBinding(() -> {
            if (openArchiveTextField.getText() == null ||
                    openArchiveTextField.getText().trim().isEmpty())
                return false;
            return new File(openArchiveTextField.getText()).exists();
        }, openArchiveTextField.textProperty());

        BooleanBinding dirBooleanBinding = Bindings.createBooleanBinding(() -> {
            if (openDirTextField.getText() == null ||
                    openDirTextField.getText().trim().isEmpty())
                return false;
            return new File(openDirTextField.getText()).isDirectory();
        }, openDirTextField.textProperty());

        BooleanBinding urlBooleanBinding = Bindings.createBooleanBinding(() -> {
            if (openURLTextField.getText() == null ||
                    openURLTextField.getText().trim().isEmpty())
                return false;
            try {
                new URL(openURLTextField.getText());
                return true;
            } catch (MalformedURLException e) {
                return false;
            }
        }, openURLTextField.textProperty());


        BooleanBinding gitBooleanBinding = Bindings.createBooleanBinding(() -> {
                    if (openGitRepoTextField.getText() == null ||
                            openGitRepoTextField.getText().trim().isEmpty())
                        return false;
                    if (useGitBranchCheckBox.isSelected() &&
                            (openGitBranchTextField.getText() == null ||
                                    openGitBranchTextField.getText().trim().isEmpty()))
                        return false;
                    if (useGitDirCheckBox.isSelected() &&
                            (openGitDirTextField.getText() == null ||
                                    openGitDirTextField.getText().trim().isEmpty()))
                        return false;
                    return true;
                }, openGitRepoTextField.textProperty(),
                useGitBranchCheckBox.selectedProperty(), openGitBranchTextField.textProperty(),
                useGitDirCheckBox.selectedProperty(), openGitDirTextField.textProperty());

        valid = Bindings.when(Bindings.equal(especSource, OpenESpecEvent.ESpecSource.ARCHIVE))
                .then(archiveBooleanBinding)
                .otherwise(
                        Bindings.when(Bindings.equal(especSource, OpenESpecEvent.ESpecSource.DIR))
                                .then(dirBooleanBinding).otherwise(
                                Bindings.when(Bindings.equal(especSource, OpenESpecEvent.ESpecSource.URL))
                                        .then(urlBooleanBinding).otherwise(
                                        Bindings.when(Bindings.equal(especSource, OpenESpecEvent.ESpecSource.GIT))
                                                .then(gitBooleanBinding).otherwise(false)
                                )
                        )
                );
    }

    @Nullable
    private static File sanitizeFolder(@Nullable String folderPath) {

        File dirFile = null;
        if (folderPath != null) {
            dirFile = new File(folderPath);
        } else
            return null;

        //try to go up the directory tree if the given file/folder does not exist (anymore)
        while (dirFile != null
                && !dirFile.exists()
                && dirFile.getParentFile() != null) {
            dirFile = dirFile.getParentFile();
        }

        //make sure that we return a folder instead of a file
        while (dirFile != null && !dirFile.isDirectory()) {
            dirFile = dirFile.getParentFile();
        }

        return dirFile;
    }


    @FXML
    public void openArchiveAction() {
        FileChooser fc = new FileChooser();
        fc.setTitle("Open Experiment Specification");

        //set the default open dir
        String initialDir = preferences.getString(GuiPreferenceKey.PREF_LAST_OPENED_ESPEC);
        if (initialDir == null && preferences.hasKey(GuiPreferenceKey.PREF_ESPEC_FILE_DIR)) {
            initialDir = preferences.getString(GuiPreferenceKey.PREF_ESPEC_FILE_DIR);
        }
        if (initialDir == null && preferences.hasKey(GuiPreferenceKey.PREF_ESPEC_DIR_DIR)) {
            initialDir = preferences.getString(GuiPreferenceKey.PREF_ESPEC_DIR_DIR);
        }
        if (initialDir == null && System.getProperty("user.home") != null) {
            initialDir = System.getProperty("user.home");
        }
        fc.setInitialDirectory(sanitizeFolder(initialDir));

        File file = fc.showOpenDialog(openArchiveTextField.getScene().getWindow());

        if (file != null) {
            openArchiveTextField.setText(file.getPath());

            if (file.getParentFile() != null) {
                preferences.setString(GuiPreferenceKey.PREF_ESPEC_FILE_DIR, file.getParentFile().getPath());
            }
        }
    }


    @FXML
    public void openDirAction() {
        DirectoryChooser dc = new DirectoryChooser();
        dc.setTitle("Open Experiment Specification");

        //set the default open dir
        String initialDir = preferences.getString(GuiPreferenceKey.PREF_LAST_OPENED_ESPEC);
        if (initialDir == null && preferences.hasKey(GuiPreferenceKey.PREF_ESPEC_DIR_DIR)) {
            initialDir = preferences.getString(GuiPreferenceKey.PREF_ESPEC_DIR_DIR);
        }
        if (initialDir == null && preferences.hasKey(GuiPreferenceKey.PREF_ESPEC_FILE_DIR)) {
            initialDir = preferences.getString(GuiPreferenceKey.PREF_ESPEC_FILE_DIR);
        }
        if (initialDir == null && System.getProperty("user.home") != null) {
            initialDir = System.getProperty("user.home");
        }
        dc.setInitialDirectory(sanitizeFolder(initialDir));

        File file = dc.showDialog(openDirTextField.getScene().getWindow());

        if (file != null) {
            openDirTextField.setText(file.getPath());
            preferences.setString(GuiPreferenceKey.PREF_ESPEC_DIR_DIR, file.getPath());
        }
    }

    private OpenESpecEvent.ESpecSource getESpecSource() {
        if (archiveSourceButton.isSelected()) return OpenESpecEvent.ESpecSource.ARCHIVE;
        if (dirSourceButton.isSelected()) return OpenESpecEvent.ESpecSource.DIR;
        if (urlSourceButton.isSelected()) return OpenESpecEvent.ESpecSource.URL;
        if (gitSourceButton.isSelected()) return OpenESpecEvent.ESpecSource.GIT;
        throw new IllegalStateException("Radiobuttons are not in an allowed state!");
    }

//    @FXML
//    public void updateValid() {
//        valid.setValue(calcValid());
//    }
//
//    private boolean calcValid() {
//
//
//        LOG.debug("Validating");
//        switch (getESpecSource()) {
//            case ARCHIVE: {
//                if (openArchiveTextField.getText() == null ||
//                        openArchiveTextField.getText().trim().isEmpty())
//                    return false;
//                return new File(openArchiveTextField.getText()).exists();
//            }
//            case DIR: {
//                if (openDirTextField.getText() == null ||
//                        openDirTextField.getText().trim().isEmpty())
//                    return false;
//                return new File(openDirTextField.getText()).isDirectory();
//            }
//            case URL: {
//                if (openURLTextField.getText() == null ||
//                        openURLTextField.getText().trim().isEmpty())
//                    return false;
//                try {
//                    new URL(openURLTextField.getText());
//                    return true;
//                } catch (MalformedURLException e) {
//                    return false;
//                }
//            }
//            case GIT: {
//                if (openGitRepoTextField.getText() == null ||
//                        openGitRepoTextField.getText().trim().isEmpty())
//                    return false;
//                if (useGitBranchCheckBox.isSelected() &&
//                        (openGitBranchTextField.getText() == null ||
//                                openGitBranchTextField.getText().trim().isEmpty()))
//                    return false;
//                if (useGitDirCheckBox.isSelected() &&
//                        (openGitDirTextField.getText() == null ||
//                                openGitDirTextField.getText().trim().isEmpty()))
//                    return false;
//                return true;
//            }
//            default: {
//                return false;
//            }
//        }
//    }
//

    public void setOnAction(EventHandler<ActionEvent> actionEventHandler) {
        this.actionEventHandler = actionEventHandler;
    }

    public boolean isValid() {
        return valid.get();
    }

    public BooleanBinding validProperty() {
        return valid;
    }

    @Nonnull
    public String getESpecSourceString() {
        switch (getESpecSource()) {
            case ARCHIVE: {
                return openArchiveTextField.getText();
            }
            case DIR: {
                return openDirTextField.getText();
            }
            case URL: {
                return openURLTextField.getText();
            }
            case GIT: {
                String res = "git ";

                if (openGitRepoTextField.getText().contains("github"))
                    res = "github ";

                res += openGitRepoTextField.getText();

                //syntax:
                //git[hub] https://<githubTLD>/<org>/<repo>.git [<subdir> [<branch>]]

                if (useGitBranchCheckBox.isSelected() || useGitDirCheckBox.isSelected()) {
                    if (useGitDirCheckBox.isSelected())
                        res += " " + openGitDirTextField.getText();
                    else
                        res += " /";
                }

                if (useGitBranchCheckBox.isSelected()) {
                    res += " " + openGitBranchTextField.getText();
                }

                return res;
            }
            default: {
                throw new IllegalStateException("Unhandled ESpec Source: " + getESpecSource());
            }
        }
    }
}
